@@ -92,6 +92,7 @@ static void uv__fast_poll_submit_poll_req(uv_loop_t* loop, uv_poll_t* handle) {
     handle->mask_events_2 = 0;
   } else {
     assert(0);
+    return;
   }
 
   /* Setting Exclusive to TRUE makes the other poll request return if there */
@@ -143,10 +144,8 @@ static int uv__fast_poll_cancel_poll_req(uv_loop_t* loop, uv_poll_t* handle) {
 
   if (result == SOCKET_ERROR) {
     DWORD error = WSAGetLastError();
-    if (error != WSA_IO_PENDING) {
-      uv__set_sys_error(loop, WSAGetLastError());
-      return -1;
-    }
+    if (error != WSA_IO_PENDING)
+      return error;
   }
 
   return 0;
@@ -168,6 +167,7 @@ static void uv__fast_poll_process_poll_req(uv_loop_t* loop, uv_poll_t* handle,
     mask_events = handle->mask_events_2;
   } else {
     assert(0);
+    return;
   }
 
   /* Report an error unless the select was just interrupted. */
@@ -175,8 +175,7 @@ static void uv__fast_poll_process_poll_req(uv_loop_t* loop, uv_poll_t* handle,
     DWORD error = GET_REQ_SOCK_ERROR(req);
     if (error != WSAEINTR && handle->events != 0) {
       handle->events = 0; /* Stop the watcher */
-      uv__set_sys_error(loop, error);
-      handle->poll_cb(handle, -1, 0);
+      handle->poll_cb(handle, uv_translate_sys_error(error), 0);
     }
 
   } else if (afd_poll_info->NumberOfHandles >= 1) {
@@ -238,17 +237,18 @@ static int uv__fast_poll_set(uv_loop_t* loop, uv_poll_t* handle, int events) {
 }
 
 
-static void uv__fast_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
+static int uv__fast_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
   handle->events = 0;
   uv__handle_closing(handle);
 
   if (handle->submitted_events_1 == 0 &&
       handle->submitted_events_2 == 0) {
     uv_want_endgame(loop, (uv_handle_t*) handle);
+    return 0;
   } else {
     /* Cancel outstanding poll requests by executing another, unique poll */
     /* request that forces the outstanding ones to return. */
-    uv__fast_poll_cancel_poll_req(loop, handle);
+    return uv__fast_poll_cancel_poll_req(loop, handle);
   }
 }
 
@@ -292,7 +292,7 @@ static SOCKET uv__fast_poll_get_peer_socket(uv_loop_t* loop,
   SOCKET peer_socket;
 
   index = -1;
-  for (i = 0; i < ARRAY_SIZE(uv_msafd_provider_ids); i++) {
+  for (i = 0; (size_t) i < ARRAY_SIZE(uv_msafd_provider_ids); i++) {
     if (memcmp((void*) &protocol_info->ProviderId,
                (void*) &uv_msafd_provider_ids[i],
                sizeof protocol_info->ProviderId) == 0) {
@@ -403,6 +403,7 @@ static void uv__slow_poll_submit_poll_req(uv_loop_t* loop, uv_poll_t* handle) {
     handle->mask_events_2 = 0;
   } else {
     assert(0);
+    return;
   }
 
   if (!QueueUserWorkItem(uv__slow_poll_thread_proc,
@@ -419,6 +420,8 @@ static void uv__slow_poll_submit_poll_req(uv_loop_t* loop, uv_poll_t* handle) {
 static void uv__slow_poll_process_poll_req(uv_loop_t* loop, uv_poll_t* handle,
     uv_req_t* req) {
   unsigned char mask_events;
+  int err;
+
   if (req == &handle->poll_req_1) {
     handle->submitted_events_1 = 0;
     mask_events = handle->mask_events_1;
@@ -427,14 +430,15 @@ static void uv__slow_poll_process_poll_req(uv_loop_t* loop, uv_poll_t* handle,
     mask_events = handle->mask_events_2;
   } else {
     assert(0);
+    return;
   }
 
   if (!REQ_SUCCESS(req)) {
     /* Error. */
     if (handle->events != 0) {
+      err = GET_REQ_ERROR(req);
       handle->events = 0; /* Stop the watcher */
-      uv__set_sys_error(loop, GET_REQ_ERROR(req));
-      handle->poll_cb(handle, -1, 0);
+      handle->poll_cb(handle, uv_translate_sys_error(err), 0);
     }
   } else {
     /* Got some events. */
@@ -477,7 +481,7 @@ static int uv__slow_poll_set(uv_loop_t* loop, uv_poll_t* handle, int events) {
 }
 
 
-static void uv__slow_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
+static int uv__slow_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
   handle->events = 0;
   uv__handle_closing(handle);
 
@@ -485,11 +489,13 @@ static void uv__slow_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
       handle->submitted_events_2 == 0) {
     uv_want_endgame(loop, (uv_handle_t*) handle);
   }
+
+  return 0;
 }
 
 
 int uv_poll_init(uv_loop_t* loop, uv_poll_t* handle, int fd) {
-  return uv_poll_init_socket(loop, handle, (SOCKET) _get_osfhandle(fd));
+  return uv_poll_init_socket(loop, handle, (SOCKET) uv__get_osfhandle(fd));
 }
 
 
@@ -499,6 +505,11 @@ int uv_poll_init_socket(uv_loop_t* loop, uv_poll_t* handle,
   int len;
   SOCKET peer_socket, base_socket;
   DWORD bytes;
+  DWORD yes = 1;
+
+  /* Set the socket to nonblocking mode */
+  if (ioctlsocket(socket, FIONBIO, &yes) == SOCKET_ERROR)
+    return uv_translate_sys_error(WSAGetLastError());
 
   /* Try to obtain a base handle for the socket. This increases this chances */
   /* that we find an AFD handle and are able to use the fast poll mechanism. */
@@ -532,8 +543,7 @@ int uv_poll_init_socket(uv_loop_t* loop, uv_poll_t* handle,
                  SO_PROTOCOL_INFOW,
                  (char*) &protocol_info,
                  &len) != 0) {
-    uv__set_sys_error(loop, WSAGetLastError());
-    return -1;
+    return uv_translate_sys_error(WSAGetLastError());
   }
 
   /* Get the peer socket that is needed to enable fast poll. If the returned */
@@ -549,7 +559,7 @@ int uv_poll_init_socket(uv_loop_t* loop, uv_poll_t* handle,
     handle->flags |= UV_HANDLE_POLL_SLOW;
   }
 
-  /* Intialize 2 poll reqs. */
+  /* Initialize 2 poll reqs. */
   handle->submitted_events_1 = 0;
   uv_req_init(loop, (uv_req_t*) &(handle->poll_req_1));
   handle->poll_req_1.type = UV_POLL_REQ;
@@ -565,12 +575,16 @@ int uv_poll_init_socket(uv_loop_t* loop, uv_poll_t* handle,
 
 
 int uv_poll_start(uv_poll_t* handle, int events, uv_poll_cb cb) {
+  int err;
+
   if (!(handle->flags & UV_HANDLE_POLL_SLOW)) {
-    if (uv__fast_poll_set(handle->loop, handle, events) < 0)
-      return -1;
+    err = uv__fast_poll_set(handle->loop, handle, events);
   } else {
-    if (uv__slow_poll_set(handle->loop, handle, events) < 0)
-      return -1;
+    err = uv__slow_poll_set(handle->loop, handle, events);
+  }
+
+  if (err) {
+    return uv_translate_sys_error(err);
   }
 
   handle->poll_cb = cb;
@@ -580,11 +594,15 @@ int uv_poll_start(uv_poll_t* handle, int events, uv_poll_cb cb) {
 
 
 int uv_poll_stop(uv_poll_t* handle) {
+  int err;
+
   if (!(handle->flags & UV_HANDLE_POLL_SLOW)) {
-    return uv__fast_poll_set(handle->loop, handle, 0);
+    err = uv__fast_poll_set(handle->loop, handle, 0);
   } else {
-    return uv__slow_poll_set(handle->loop, handle, 0);
+    err = uv__slow_poll_set(handle->loop, handle, 0);
   }
+
+  return uv_translate_sys_error(err);
 }
 
 
@@ -597,11 +615,11 @@ void uv_process_poll_req(uv_loop_t* loop, uv_poll_t* handle, uv_req_t* req) {
 }
 
 
-void uv_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
+int uv_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
   if (!(handle->flags & UV_HANDLE_POLL_SLOW)) {
-    uv__fast_poll_close(loop, handle);
+    return uv__fast_poll_close(loop, handle);
   } else {
-    uv__slow_poll_close(loop, handle);
+    return uv__slow_poll_close(loop, handle);
   }
 }
 
